# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import argparse
import sys

from .backupman import BackupManager
from .utils import getString


class ArgparseCustomFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text[:2] == 'F!':
            return text.splitlines()[1:]
        return argparse.HelpFormatter._split_lines(self, text, width)


def main() -> int:
    # create CLI and parse arguments with argparse
    parser = argparse.ArgumentParser(description=getString("BackuPy: A simple backup program in python with an emphasis on data integrity and transparent behaviour"),
                                     formatter_class=lambda prog: ArgparseCustomFormatter(prog, max_help_position=15),
                                     usage="%(prog)s [options] -- <source> <dest>\n"
                                           "       %(prog)s <source> <dest> [options]\n"
                                           "       %(prog)s <source> --load [-c mode] [--dbscan] [--dry-run]\n"
                                           "       %(prog)s -h | --help")
    parser.add_argument("source", action="store", type=str,
                        help=getString("Path to source"))
    parser.add_argument("dest", action="store", type=str, nargs="?", default=None,
                        help=getString("Path to destination"))
    group1 = parser.add_argument_group("file mode options", "")
    group2 = parser.add_argument_group("misc file options", "")
    group3 = parser.add_argument_group("execution options", "")
    group4 = parser.add_argument_group("configuration options", "")
    group1.add_argument("-m", type=str.lower, dest="main_mode", default="mirror", metavar="mode", choices=["mirror", "backup", "sync"],
                        help=getString(
                             "F!\n"
                             "Main mode: for files that exist only on one side\n"
                             "  MIRROR (default)\n"
                             "    [source-only -> destination, delete destination-only]\n"
                             "  BACKUP\n"
                             "    [source-only -> destination, keep destination-only]\n"
                             "  SYNC\n"
                             "    [source-only -> destination, destination-only -> source]"))
    group1.add_argument("-s", type=str.lower, dest="select_mode", default="source", metavar="mode", choices=["source", "dest", "new", "no"],
                        help=getString(
                             "F!\n"
                             "Selection mode: for files that exist on both sides but differ\n"
                             "  SOURCE (default)\n"
                             "    [copy source to destination]\n"
                             "  DEST\n"
                             "    [copy destination to source]\n"
                             "  NEW\n"
                             "    [copy newer to opposite side]\n"
                             "  NO\n"
                             "    [do nothing]"))
    group1.add_argument("-c", type=str.lower, dest="compare_mode", default=None, metavar="mode", choices=["attr", "attr+", "crc"],
                        help=getString(
                             "F!\n"
                             "Compare mode: for detecting which files differ\n"
                             "  ATTR (default)\n"
                             "    [compare file attributes: mod-time and size]\n"
                             "  ATTR+\n"
                             "    [compare file attributes and record CRC for changed files]\n"
                             "  CRC\n"
                             "    [compare file attributes and CRC for every file]"))
    group2.add_argument("--sync-delete", dest="sync_propagate_deletions", action="store_true",
                        help=getString("Use the database to propagate deletions since the last sync"))
    group2.add_argument("--fi", dest="filter_include_list", action="store", type=str, nargs="+", default=None, metavar="regex",
                        help=getString("Filter: Only include files matching the regular expression(s) (include all by default, searches file paths)"))
    group2.add_argument("--fe", dest="filter_exclude_list", action="store", type=str, nargs="+", default=None, metavar="regex",
                        help=getString("Filter: Exclude files matching the regular expression(s) (exclude has priority over include, searches file paths)"))
    group2.add_argument("--noarchive", dest="noarchive", action="store_true",
                        help=getString(
                             "F!\n"
                             "Disable archiving files before overwriting/deleting to:\n"
                             "   <source|dest>/.backupy/Archives/yymmdd-HHMM/\n"
                             "   <source|dest>/.backupy/Trash/yymmdd-HHMM/"))
    group2.add_argument("--nomoves", dest="nomoves", action="store_true",
                        help=getString("Do not detect when files are moved or renamed"))
    group3.add_argument("--noprompt", dest="noprompt", action="store_true",
                        help=getString("Complete run without prompting for confirmation"))
    group3.add_argument("-d", "--dbscan", dest="scan_only", action="store_true",
                        help=getString("Only scan files to check and update their database entries"))
    group3.add_argument("-n", "--dry-run", dest="dry_run", action="store_true",
                        help=getString("Perform a dry run with no changes made to your files"))
    group3.add_argument("-q", "--qconflicts", dest="quit_on_db_conflict", action="store_true",
                        help=getString(
                             "F!\n"
                             "Quit if database conflicts are detected (always notified)\n"
                             "  -> unexpected changes on destination (backup and mirror)\n"
                             "  -> sync conflict (file modified on both sides since last sync)\n"
                             "  -> file corruption (ATTR+ or CRC compare modes)"))
    group3.add_argument("-v", "--verify", dest="verify_copy", action="store_true",
                        help=getString("Verify CRC of copied files"))
    group4.add_argument("--nolog", dest="nolog", action="store_true",
                        help=getString(
                             "F!\n"
                             "Disable writing log and file databases to:\n"
                             "   <source>/.backupy/Logs/log-yymmdd-HHMM.csv\n"
                             "   <source|dest>/.backupy/database.json"))
    group4.add_argument("-p", "--posix", dest="force_posix_path_sep", action="store_true",
                        help=getString("Force posix style paths on non-posix operating systems"))
    group4.add_argument("-k", "--save", dest="save", action="store_true",
                        help=getString("Save configuration to <source>/.backupy/config.json"))
    group4.add_argument("-l", "--load", dest="load", action="store_true",
                        help=getString("Load configuration from <source>/.backupy/config.json"))
    parser.add_argument("--nocolour", dest="nocolour", action="store_true",
                        help=argparse.SUPPRESS)
    args = parser.parse_args()
    # create and run job
    backup_manager = BackupManager(args)
    return backup_manager.run()


if __name__ == "__main__":
    # execute with python -m backupy.cli
    sys.exit(main())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import os
import shutil
import time

from .config import ConfigObject
from .dirinfo import DirInfo
from .utils import (
    getString,
    writeCsv,
)


class LogManager:
    def __init__(self, backup_time: int, gui: bool):
        """Provides methods for log formatting and pretty printing (used by BackupManager)"""
        # init variables
        self._log = []
        self._log_columns = []
        self._log_row_split = False
        self.backup_time = backup_time
        self.gui = gui
        self.terminal_width = shutil.get_terminal_size()[0]
        # gui modifications
        if self.gui:
            from .gui import colourize
            self.gui_colourize = colourize
            self.terminal_width = 80
        # init attributes for linting (replaced by BackupManager to reference the same object)
        self.config = ConfigObject
        self.source = DirInfo
        self.dest = DirInfo

    def append(self, entry: list, columns: list = [], row_split: bool = True) -> None:
        # entry to appear in log, label columns for summary, row_split for a new file
        # columns corresponds to the column names of each item in entry or an empty string to omit from summary
        # columns should be an empty list if this entry should be omitted entirely from summary
        # len(columns) should be >= len(entry) or an empty list
        self._log.append(entry)
        if columns:
            self._log_columns.append(columns + [row_split or self._log_row_split])
            self._log_row_split = False
        else:
            self._log_columns.append([])

    def appendNewRowFlag(self) -> None:
        self._log_row_split = True

    def convertLog(self) -> list:
        # convert existing log into log summary
        # combines multiple log entries into a single row
        # each row contains all relevant info of a single file across source, destination and databases on both sides
        columns = ["Section",
                   "Header", "Subheader", "Path", "Size", "Modified", "Hash",
                   "Header", "Subheader", "Path", "Size", "Modified", "Hash",
                   "Header", "Subheader", "Path", "Size", "Modified", "Hash",
                   "Header", "Subheader", "Path", "Size", "Modified", "Hash"]
        log_csv = [columns]
        new_section = False
        for i in range(len(self._log)):
            if self._log_columns[i]:
                if self._log_columns[i][-1] and not new_section:
                    # create new row based on last value of log_columns (unless a new section was just started)
                    log_csv.append([""]*len(columns))
                    entry_number = -1
                if self._log_columns[i][0] == "Section":
                    # fill in the label for the new section
                    log_csv[-1][0] = self._log[i][0].strip("# ")
                    new_section = True
                    entry_number = -1
                else:
                    new_entry = True
                    for j in range(len(self._log[i])):
                        # fill in row with items from entries under the correct column
                        if self._log_columns[i][j]:
                            if new_entry:
                                entry_number += 1
                                new_entry = False
                                new_section = False
                            entry_offset = columns.index(self._log_columns[i][j])
                            log_csv[-1][entry_offset + entry_number*6] = self._log[i][j]
                    if not log_csv[-1][0]:
                        # copy the section from the above row if missing
                        log_csv[-1][0] = log_csv[-2][0]
        if all(entry == "" for entry in log_csv[-1][1:]):
            _ = log_csv.pop(-1)
        return log_csv

    def writeLog(self, db_name: str) -> None:
        if not self.config.nolog:
            # <source|dest>/.backupy/database.json
            if self.config.dry_run:
                db_name = db_name[:-4] + "dryrun.json"
            self.source.saveDatabase(db_name)
            self.dest.saveDatabase(db_name)
            self._log[1][5] = self.source.calcCrc(os.path.join(self.source.dir, self.source.config_dir, db_name))
            self._log[1][7] = self.dest.calcCrc(os.path.join(self.dest.dir, self.dest.config_dir, db_name))
            # <source>/.backupy/Logs/log-yymmdd-HHMM.csv
            if self.config.root_alias_log or self.config.force_posix_path_sep:
                for i in range(2, len(self._log)):
                    for j in range(len(self._log[i])):
                        if type(self._log[i][j]) == str:
                            if self.config.root_alias_log:
                                self._log[i][j] = self._log[i][j].replace(self.config.source, getString("<source>"))
                                self._log[i][j] = self._log[i][j].replace(self.config.dest, getString("<dest>"))
                            if self.config.force_posix_path_sep:
                                self._log[i][j] = self._log[i][j].replace(os.path.sep, "/")
            writeCsv(os.path.join(self.config.source, self.config.log_dir, "log-" + self.backup_time + ".csv"), self._log)
            if self.config.write_log_dest:
                writeCsv(os.path.join(self.config.dest, self.config.log_dir, "log-" + self.backup_time + "-dest.csv"), self._log)
            if self.config.write_log_summary:
                writeCsv(os.path.join(self.config.source, self.config.log_dir, "log-" + self.backup_time + "-plus.csv"), self.convertLog())

    def replaceSurrogates(self, string: str) -> str:
        return string.encode("utf-8", "surrogateescape").decode("utf-8", "replace")

    def colourString(self, string: str, colour: str) -> str:
        string = self.replaceSurrogates(string)
        if self.gui:
            return self.gui_colourize(string, colour)
        elif self.config.nocolour:
            return string
        colours = {
            "R": "\033[91m",
            "G": "\033[92m",
            "Y": "\033[93m",
            "B": "\033[94m",
            "V": "\033[95m"
        }
        return colours[colour] + string + "\033[0m"

    def prettySize(self, size: float) -> str:
        if size > 10**9:
            return "{:<10}".format("%s GB" % (round(size/10**9, 2)))
        elif size > 10**6:
            return "{:<10}".format("%s MB" % (round(size/10**6, 2)))
        elif size > 10**3:
            return "{:<10}".format("%s kB" % (round(size/10**3, 2)))
        else:
            return "{:<10}".format("%s B" % (size))

    def prettyAttr(self, attr: dict) -> list:
        attr_list = []
        attr_list.append(self.prettySize(attr["size"]).strip())
        attr_list.append(time.ctime(attr["mtime"]))
        if "crc" in attr:
            attr_list.append(attr["crc"])
        if "dir" in attr:
            attr_list.append("dir: %s" % (attr["dir"]))
        return attr_list

    def colourPrint(self, msg: str, colour: str) -> None:
        if self.config.verbose:
            if colour == "NONE":
                print(self.replaceSurrogates(msg))
            else:
                print(self.colourString(msg, colour))

    def printFileInfo(self, header: str, f: str, d: dict, sub_header: str = "", skip_info: bool = False) -> None:
        header, sub_header = getString(header), getString(sub_header)
        if f in d and d[f] is not None:
            self.append([header.strip(), sub_header.strip(), f] + self.prettyAttr(d[f]), ["Header", "Subheader", "Path", "Size", "Modified", "Hash", ""], False)
            missing = False
        else:
            self.append([header.strip(), sub_header.strip(), f] + [getString("Missing")], ["Header", "Subheader", "Path", "Size"], False)
            missing = True
        if header == "":
            s = ""
        else:
            s = self.colourString(header, "B") + self.replaceSurrogates(f)
            if not skip_info:
                s = s + "\n"
        if not skip_info:
            extra_space = " "*min(4, self.terminal_width//5-16)
            s = s + extra_space + self.colourString(sub_header, "B") + " "*(12-len(sub_header))
            if not missing:
                s = s + extra_space + self.colourString(getString(" Size: "), "B") + self.prettySize(d[f]["size"])
                s = s + extra_space + self.colourString(getString(" Modified: "), "B") + time.ctime(d[f]["mtime"])
                if "crc" in d[f]:
                    s = s + extra_space + self.colourString(getString(" Hash: "), "B") + d[f]["crc"]
            else:
                s = s + extra_space + self.colourString(getString(" Missing"), "B")
        print(s)

    def printFiles(self, files: list, d: dict) -> None:
        for f in files:
            self.appendNewRowFlag()
            self.printFileInfo("File: ", f, d)

    def printChangedFiles(self, files: list, d1: dict, d2: dict, s1: str = " Source", s2: str = "   Dest") -> None:
        for f in files:
            self.appendNewRowFlag()
            self.printFileInfo("File: ", f, d1, " "*4 + s1)
            self.printFileInfo("", f, d2, " "*4 + s2)

    def printMovedFiles(self, files: list, d1: dict, d2: dict, h1: str = "Source: ", h2: str = "  Dest: ") -> None:
        for f in files:
            self.appendNewRowFlag()
            if self.config.main_mode != "sync":
                self.printFileInfo(h1, f["source"], d1, skip_info=True)
                self.printFileInfo(h2, f["dest"], d2)
            elif f["match"] == "source":
                self.printFileInfo(h1, f["source"], d1, skip_info=True)
                self.printFileInfo(h2, f["dest"], d2, "Move Dest  ")
            elif f["match"] == "dest":
                self.printFileInfo(h1, f["source"], d1, skip_info=True)
                self.printFileInfo(h2, f["dest"], d2, "Move Source")

    def printSyncDbConflicts(self, files: list, d1: dict, d2: dict, d1db: dict, d2db: dict) -> None:
        for f in files:
            self.appendNewRowFlag()
            self.printFileInfo("File: ", f, d1, "     Source")
            self.printFileInfo("", f, d1db, "         DB")
            self.printFileInfo("", f, d2, "       Dest")
            self.printFileInfo("", f, d2db, "         DB")



from .backupman import BackupManager
from .cli import main
from .utils import getVersion

__all__ = ["create_job", "run", "start_gui"]


def create_job(config: dict) -> BackupManager:
    """Create a new job for a given configuration, returns a BackupManager object"""
    return BackupManager(config)


def run(config: dict) -> int:
    """Execute backupy for a given configuration"""
    backup_man = BackupManager(config)
    return backup_man.run()


def start_gui():
    """Launch the GUI"""
    from .gui import main_gui
    main_gui()



import os
import sys
import tempfile
sys.stdout = tempfile.TemporaryFile()

import PySimpleGUI as sg
from colored import stylize, attr, fg
from gooey import Gooey, GooeyParser, local_resource_path

from .backupman import BackupManager
from .utils import (
    getVersion,
    readJson,
    writeJson,
)


class Unbuffered(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


sys.stdout = Unbuffered(sys.__stdout__)

GPLv3 = """This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>."""


def colourize(string: str, colour: str) -> str:
    colours = {
            "R": fg("red"),
            "G": fg("green"),
            "B": fg("blue"),
            "Y": fg("yellow"),
            "V": fg("magenta")
        }
    return stylize(string, colours[colour])


def simplePrompt(msg: str) -> str:
    sg.change_look_and_feel("System Default For Real")
    layout = [[sg.Text(msg)],
              [sg.Button("Ok"), sg.Button("Cancel")]]
    window = sg.Window("BackuPy", layout)
    event, _ = window.Read()
    window.close()
    if event == "Ok":
        return "y"
    else:
        return "n"


@Gooey(program_name="BackuPy",
       image_dir=local_resource_path("images/"),
       richtext_controls=True,
       tabbed_groups=True,
       monospace_display=True,
       progress_regex=r"^progress: (?P<current>\d+)/(?P<total>\d+)$",
       progress_expr="current / total * 100",
       hide_progress_msg=True,
       default_size=(620, 530),
       menu=[{
            'name': 'File',
            'items': [{
                    'type': 'AboutDialog',
                    'menuTitle': 'About',
                    'name': 'BackuPy',
                    'description': 'A simple backup program in python',
                    'version': getVersion(),
                    'website': 'https://github.com/elesiuta/backupy',
                    'license': GPLv3
            }]
        }])
def main_gui():
    # load profiles
    dict_profiles = readJson("profiles.json")
    if "profiles" in dict_profiles:
        list_profiles = dict_profiles["profiles"]
    else:
        list_profiles = []
    # argparse setup
    parser = GooeyParser(description="A simple backup program in python")
    group1 = parser.add_argument_group("Profiles", "")
    group2 = parser.add_argument_group("Directories", "")
    group3 = parser.add_argument_group("Usage Mode", "")
    group4 = parser.add_argument_group("Configuration", "")
    for source_dir in list_profiles:
        group1.add_argument("--load_profile_"+source_dir, action="store_true", metavar="Load: "+os.path.basename(source_dir),
                            gooey_options={"full_width": True, "show_label": True},
                            help=" " + source_dir)
    group2.add_argument("--source", metavar="Source", action="store", type=str, default=None,
                        widget="DirChooser", gooey_options={"full_width": True},
                        help="Path of source")
    group2.add_argument("--dest", metavar="Destination", action="store", type=str, default=None,
                        widget="DirChooser", gooey_options={"full_width": True},
                        help="Path of destination")
    group3_main = group3.add_mutually_exclusive_group(required=True,
                                                      gooey_options={"title": "Main mode: How to handle files that exist only on one side?",
                                                                     "full_width": True,
                                                                     'initial_selection': 0})
    group3_main.add_argument("--main_mode_radio_mirror", metavar="Mirror", action="store_true",
                             help="[source-only -> destination, delete destination-only]")
    group3_main.add_argument("--main_mode_radio_backup", metavar="Backup", action="store_true",
                             help="[source-only -> destination, keep destination-only]")
    group3_main.add_argument("--main_mode_radio_sync", metavar="Sync", action="store_true",
                             help="[source-only -> destination, destination-only -> source]")
    group3_select = group3.add_mutually_exclusive_group(required=True,
                                                        gooey_options={"title": "Selection mode: How to handle files that exist on both sides but differ?",
                                                                       "full_width": True,
                                                                       'initial_selection': 0})
    group3_select.add_argument("--select_mode_radio_source", metavar="Source", action="store_true",
                               help="[copy source to destination]")
    group3_select.add_argument("--select_mode_radio_dest", metavar="Destination", action="store_true",
                               help="[copy destination to source]")
    group3_select.add_argument("--select_mode_radio_new", metavar="New", action="store_true",
                               help="[copy newer to opposite side]")
    group3_select.add_argument("--select_mode_radio_no", metavar="None", action="store_true",
                               help="[do nothing]")
    group3_compare = group3.add_mutually_exclusive_group(required=True,
                                                         gooey_options={"title": "Compare mode: How to detect files that exist on both sides but differ?",
                                                                        "full_width": True,
                                                                        'initial_selection': 0})
    group3_compare.add_argument("--compare_mode_radio_attr", metavar="Attribute", action="store_true",
                                help="[compare file attributes: mod-time and size]")
    group3_compare.add_argument("--compare_mode_radio_attr+", metavar="Attribute+", action="store_true",
                                help="[compare file attributes and record CRC for changed files]")
    group3_compare.add_argument("--compare_mode_radio_crc", metavar="CRC", action="store_true",
                                help="[compare file attributes and CRC for every file]")
    group4.add_argument("--dry-run", dest="dry_run", action="store_true", metavar="Dry run", gooey_options={"full_width": True},
                        help=" Perform a dry run with no changes made to your files")
    group4.add_argument("-p", "--posix", dest="--force_posix_path_sep", action="store_true", metavar="Force posix paths", gooey_options={"full_width": True},
                        help=" Force posix style paths on non-posix operating systems")
    group4.add_argument("--sync-delete", dest="sync_propagate_deletions", action="store_true", metavar="Propagate deletions during sync", gooey_options={"full_width": True},
                        help=" Use the database to propagate deletions since the last sync")
    group4.add_argument("--noarchive", action="store_true", metavar="No archiving", gooey_options={"full_width": True},
                        help=" Disable archiving files before overwriting/deleting to:\n"
                             "  <source|dest>/.backupy/Archives/yymmdd-HHMM/\n"
                             "  <source|dest>/.backupy/Trash/yymmdd-HHMM/")
    group4.add_argument("--nolog", action="store_true", metavar="No logs", gooey_options={"full_width": True},
                        help=" Disable writing log and file databases to:\n"
                             "  <source>/.backupy/Logs/log-yymmdd-HHMM.csv\n"
                             "  <source|dest>/.backupy/database.json")
    group4.add_argument("--nomoves", action="store_true", metavar="No moves", gooey_options={"full_width": True},
                        help=" Do not detect when files are moved or renamed")
    group4.add_argument("--noprompt", action="store_true", metavar="No prompt", gooey_options={"full_width": True},
                        help=" Complete run without prompting for confirmation")
    group4.add_argument("-q", "--qconflicts", dest="--quit_on_db_conflict", action="store_true", metavar="Quit on database conflicts", gooey_options={"full_width": True},
                        help=" Quit if database conflicts are detected (always notified)\n"
                             "  -> unexpected changes on destination (backup and mirror)\n"
                             "  -> sync conflict (file modified on both sides since last sync)\n"
                             "  -> file corruption (ATTR+ or CRC compare modes)")
    group4.add_argument("-d", "--dbscan", dest="scan_only", action="store_true", metavar="Scan only", gooey_options={"full_width": True},
                        help=" Only scan files to check and update their database entries")
    group4.add_argument("-v", "--verify", dest="verify_copy", action="store_true", metavar="Verify copies", gooey_options={"full_width": True},
                        help=" Verify CRC of copied files")
    group2.add_argument("--save", action="store_true", metavar="Save",
                        help=" Save configuration to <source>/.backupy/config.json")
    group2.add_argument("--load", action="store_true", metavar="Load",
                        help=" Load configuration from <source>/.backupy/config.json")
    # parse args and store dictionary
    args = vars(parser.parse_args())
    # convert radio groups back to choice of string
    for key in list(args.keys()):
        if "_radio_" in key:
            if args[key] is True:
                key_split = key.split("_radio_")
                args[key_split[0]] = key_split[1]
    # check for loaded profiles
    loaded_profiles = []
    for key in list(args.keys()):
        if "load_profile_" in key:
            if args[key] is True:
                loaded_profiles.append(key.lstrip("load_profile_"))
    # execute selected profiles or execute config
    if len(loaded_profiles) >= 1:
        for source_dir in loaded_profiles:
            args["source"] = source_dir
            args["load"] = True
            backup_manager = BackupManager(args, gui=True)
            backup_manager.run()
            print("")
    else:
        # store profile if new
        if (args["save"] or args["load"]) and args["source"] not in list_profiles:
            list_profiles.append(args["source"])
            writeJson("profiles.json", {"profiles": list_profiles}, False)
        # check config and execute
        if args["source"] is not None and (args["dest"] is not None or args["load"]):
            backup_manager = BackupManager(args, gui=True)
            backup_manager.run()
            print("")
        else:
            print(colourize("At least one of the following conditions must be satisfied:\n"
                            "\t1) Select at least one profile\n"
                            "\t2) Specify source and destination directories\n"
                            "\t3) Specify source directory and load configuration\n", "R"))


if __name__ == "__main__":
    # execute with python -m backupy.gui
    sys.exit(main_gui())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import os
import random


class ConfigObject:
    def __init__(self, config: dict):
        """Used for storing user configuration, use these attribute names as keys in your configuration dictionary"""
        # default config (from argparse)
        self.source = None
        self.dest = None
        self.main_mode = "mirror"
        self.select_mode = "source"
        self.compare_mode = "attr"
        self.sync_propagate_deletions = False
        self.filter_include_list = None
        self.filter_exclude_list = None
        self.noarchive = False
        self.nocolour = False
        self.nolog = False
        self.nomoves = False
        self.noprompt = False
        self.dry_run = False
        self.force_posix_path_sep = False
        self.quit_on_db_conflict = False
        self.scan_only = False
        self.verify_copy = False
        # default config (additional)
        self.source_unique_id = "%05x" % random.randrange(16**5)
        self.dest_unique_id = "%05x" % random.randrange(16**5)
        self.archive_dir = ".backupy/Archive"
        self.config_dir = ".backupy"
        self.log_dir = ".backupy/Logs"
        self.trash_dir = ".backupy/Trash"
        self.cleanup_empty_dirs = True
        self.root_alias_log = True
        self.stdout_status_bar = True
        self.verbose = True
        self.write_database_x2 = False
        self.write_log_dest = False
        self.write_log_summary = False
        # load config
        for key in config:
            if config[key] is not None and hasattr(self, key):
                self.__setattr__(key, config[key])
        # normalize paths (these should be relative, not absolute!)
        self.archive_dir = os.path.normpath(self.archive_dir)
        self.config_dir = os.path.normpath(self.config_dir)
        self.log_dir = os.path.normpath(self.log_dir)
        self.trash_dir = os.path.normpath(self.trash_dir)

    def __setattr__(self, name, value):
        if not hasattr(self, "locked") or not self.locked:
            super().__setattr__(name, value)
        else:
            raise Exception("Error: Config modified during run (should be locked)")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import shutil
import unicodedata

from .utils import getString


class StatusBar:
    def __init__(self, title: str, total: int, display: bool, gui: bool = False):
        self.title = title
        self.total = total
        self.display = display
        self.gui = gui
        terminal_width = shutil.get_terminal_size()[0]
        if terminal_width < 16:
            self.display = False
        if self.display:
            self.char_display = terminal_width - 2
            self.progress = 0
            if self.total == -1:
                progress_str = str(self.progress) + ": "
            else:
                self.digits = str(len(str(self.total)))
                progress_str = str("{:>" + self.digits + "}").format(self.progress) + "/" + str(self.total) + ": "
            self.title_str = getString(self.title) + " "
            self.msg_len = self.char_display - len(progress_str) - len(self.title_str)
            msg = " " * self.msg_len
            print(self.title_str + progress_str + msg, end="\r")
        elif self.gui and self.total > 0:
            self.progress = 0
            print("progress: %s/%s" % (self.progress, self.total))

    def getStringMaxWidth(self, string: str) -> int:
        width = 0
        for char in string:
            if unicodedata.east_asian_width(char) in ["W", "F", "A"]:
                width += 2
            else:
                width += 1
        return width

    def update(self, msg: str) -> None:
        if self.display:
            self.progress += 1
            msg = msg.encode("utf-8", "surrogateescape").decode("utf-8", "replace")
            if self.total == -1:
                progress_str = str(self.progress) + ": "
            else:
                progress_str = str("{:>" + self.digits + "}").format(self.progress) + "/" + str(self.total) + ": "
            self.msg_len = self.char_display - len(progress_str) - len(self.title_str)
            while self.getStringMaxWidth(msg) > self.msg_len:
                splice = (len(msg) - 4) // 2
                msg = msg[:splice] + "..." + msg[-splice:]
            msg = msg + " " * int(self.msg_len - self.getStringMaxWidth(msg))
            print(self.title_str + progress_str + msg, end="\r")
        elif self.gui and self.total > 0:
            self.progress += 1
            print("progress: %s/%s" % (self.progress, self.total))

    def endProgress(self) -> None:
        if self.display:
            if self.title == "Copying":
                title_str = getString("File operations completed!")
            else:
                title_str = getString(self.title + " completed!")
            # if self.total == 0:
            #     title_str = "No action necessary"
            print(title_str + " " * (self.char_display - len(title_str)))
        elif self.gui and self.total > 0:
            self.progress = self.total
            print("progress: %s/%s" % (self.progress, self.total))
        self.display, self.gui = False, False



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import os
import re
import typing
import zlib

from .config import ConfigObject
from .statusbar import StatusBar
from .utils import (
    readJson,
    writeJson,
)


class DirInfo:
    def __init__(self, directory_root_path: str, unique_id: str,
                 other_root_path: str, config: ConfigObject, gui: bool = False):
        """For scanning directories, tracking files and changes, meant only for internal use by BackupManager"""
        # File dictionaries, keys are paths relative to directory_root_path, values are dictionaries of file attributes
        self.dict_current = {}
        self.dict_prev = {}
        # File sets of relative paths
        self.set_unmodified = set()
        self.set_modified = set()
        self.set_missing = set()
        self.set_new = set()
        self.set_crc_errors = set()
        self.set_dirs = set()
        # Init filters
        self.filter_include_list = None
        self.filter_exclude_list = None
        try:
            if config.filter_include_list is not None:
                self.filter_include_list = [re.compile(f) for f in config.filter_include_list]
            if config.filter_exclude_list is not None:
                self.filter_exclude_list = [re.compile(f) for f in config.filter_exclude_list]
        except Exception:
            raise Exception("Filter Processing Error")
        # Init variables from config
        self.compare_mode = config.compare_mode
        self.config_dir = config.config_dir
        self.ignored_toplevel_folders = list(set([config.archive_dir, config.log_dir, config.trash_dir, config.config_dir]))
        self.force_posix_path_sep = config.force_posix_path_sep
        self.write_database_x2 = config.write_database_x2 and not config.scan_only
        # Init other variables
        self.dir = directory_root_path
        self.other_dir = other_root_path
        self.unique_id = unique_id
        self.gui = gui

    def getDicts(self) -> tuple:
        """Returns tuple of dictionaries: current, prev"""
        return (self.dict_current,
                self.dict_prev)

    def getSets(self) -> tuple:
        """Returns tuple of sets: new, modified, missing, crc_errors, dirs, unmodified"""
        return (self.set_new,
                self.set_modified,
                self.set_missing,
                self.set_crc_errors,
                self.set_dirs,
                self.set_unmodified)

    def saveDatabase(self, db_name: str = "database.json") -> None:
        """Write database to config_dir on self and other if enabled"""
        writeJson(os.path.join(self.dir, self.config_dir, db_name), self.dict_current, sort_keys=True)
        if self.write_database_x2:
            other_db_path = os.path.join(self.other_dir, self.config_dir, "database-%s%s" % (self.unique_id, db_name[8:]))
            writeJson(other_db_path, self.dict_current, sort_keys=True)

    def loadDatabase(self) -> None:
        """Load database from config_dir"""
        self.dict_prev = readJson(os.path.join(self.dir, self.config_dir, "database.json"))

    def getDatabaseX2(self) -> dict:
        """Get the 'last seen' database of this directory from the perspective of the other directory"""
        other_db_path = os.path.join(self.other_dir, self.config_dir, "database-%s.json" % self.unique_id)
        database_x2 = readJson(other_db_path)
        if database_x2:
            return database_x2
        else:
            return self.dict_prev

    def verifyCrcOnCopy(self, source_root: str, dest_root: str, source_file: str, dest_file: str, secondInfo: 'DirInfo') -> None:
        if self.dir == source_root and secondInfo.dir == dest_root:
            if secondInfo.getCrc(dest_file, recalc=True) != self.getCrc(source_file):
                raise Exception("CRC Verification Failed")
        elif self.dir == dest_root and secondInfo.dir == source_root:
            if self.getCrc(dest_file, recalc=True) != secondInfo.getCrc(source_file):
                raise Exception("CRC Verification Failed")

    def updateDictOnCopy(self, source_root: str, dest_root: str, source_file: str, dest_file: str, secondInfo: 'DirInfo') -> None:
        if self.dir == source_root and secondInfo.dir == dest_root:
            secondInfo.dict_current[dest_file] = self.dict_current[source_file].copy()
        elif self.dir == dest_root and secondInfo.dir == source_root:
            self.dict_current[dest_file] = secondInfo.dict_current[source_file].copy()
        else:
            raise Exception("Update Database Error")

    def updateDictOnMove(self, source_root: str, dest_root: str, source_file: str, dest_file: str, secondInfo: 'DirInfo') -> None:
        if source_root == dest_root == self.dir:
            self.dict_current[dest_file] = self.dict_current.pop(source_file)
        elif source_root == dest_root == secondInfo.dir:
            secondInfo.dict_current[dest_file] = secondInfo.dict_current.pop(source_file)
        elif source_root == self.dir and dest_root != secondInfo.dir:
            _ = self.dict_current.pop(source_file)
        elif source_root == secondInfo.dir and dest_root != self.dir:
            _ = secondInfo.dict_current.pop(source_file)
        else:
            raise Exception("Update Database Error")

    def updateDictOnRemove(self, root_path: str, file_relative_path: str, secondInfo: 'DirInfo') -> None:
        if root_path == self.dir:
            _ = self.dict_current.pop(file_relative_path)
        elif root_path == secondInfo.dir:
            _ = secondInfo.dict_current.pop(file_relative_path)
        else:
            raise Exception("Update Database Error")

    def getCrc(self, relative_path: str, recalc: bool = False) -> str:
        if relative_path not in self.dict_current:
            self.dict_current[relative_path] = {"size": 0, "mtime": 0}
        if recalc or "crc" not in self.dict_current[relative_path]:
            full_path = os.path.join(self.dir, relative_path)
            self.dict_current[relative_path]["crc"] = self.calcCrc(full_path)
        return self.dict_current[relative_path]["crc"]

    def calcCrc(self, file_path: str, prev: int = 0) -> str:
        try:
            with open(file_path, "rb") as f:
                for line in f:
                    prev = zlib.crc32(line, prev)
            return "%X" % (prev & 0xFFFFFFFF)
        except Exception:
            # file either removed by user, or another program such as antimalware (using realtime monitoring) during scan, or lack permissions
            raise Exception("Exiting, error trying to read file: " + file_path)

    def timeMatch(self, t1: float, t2: float, exact_only: bool = False, tz_diffs: list = [3600, 3601, 3602], fs_tol: int = 2) -> bool:
        if t1 == t2:
            return True
        elif exact_only:
            return False
        diff = abs(int(t1) - int(t2))
        if diff <= fs_tol or diff in tz_diffs:
            return True
        else:
            return False

    def pathMatch(self, path: str, path_list: list) -> bool:
        # is path in path_list (or a subdir of one in it)
        if os.path.isabs(path):
            relpath, _abspath = os.path.relpath(path, self.dir), path
        else:
            relpath, _abspath = path, os.path.join(self.dir, path)
        for p in path_list:
            p = os.path.normcase(p)
            if os.path.isabs(p):
                raise Exception("Default .backupy dirs have been changed to absolute paths in the config, they should be relative paths.")
                # if os.path.normcase(os.path.commonpath([p, abspath])) == p:
                #     return True
            else:
                if os.path.normcase(os.path.commonpath([p, relpath])) == p:
                    return True
        return False

    def fileMatch(self, f1: str, f2: str, other_db: dict, other_crc_errors: set, exact_time: bool) -> bool:
        if self.dict_current[f1]["size"] == other_db[f2]["size"]:
            if self.timeMatch(self.dict_current[f1]["mtime"], other_db[f2]["mtime"], exact_time):
                # unchanged files (probably)
                if "crc" in self.dict_current[f1] and "crc" in other_db[f2] and self.dict_current[f1]["crc"] != other_db[f2]["crc"]:
                    # size and date match, but crc does not, probably corrupted, log error if scanning or comparing sides (f1 == f2), otherwise this is just checking if moved (f1 != f2) (but if time is exact, still flag it to be safe)
                    if f1 == f2 or exact_time:
                        self.set_crc_errors.add(f1)
                        other_crc_errors.add(f2)
                    return False
                return True
        return False

    def scanDir(self, stdout_status_bar: bool) -> None:
        # init
        if os.path.isdir(self.dir) and self.dict_current == {}:
            total = sum(len(f) for r, d, f in os.walk(self.dir))
            scan_status = StatusBar("Scanning", total, stdout_status_bar, gui=self.gui)
            # will never enable followlinks, adds too many possible issues and complexity in handling them
            for dir_path, subdir_list, file_list in os.walk(self.dir, followlinks=False):
                # ignore folders
                if self.pathMatch(dir_path, self.ignored_toplevel_folders):
                    subdir_list.clear()
                    continue
                # apply filters
                if self.filter_include_list is not None:
                    subdir_list = filter(lambda x: any([True if r.search(x) else False for r in self.filter_include_list]), subdir_list)
                    file_list = filter(lambda x: any([True if r.search(x) else False for r in self.filter_include_list]), file_list)
                if self.filter_exclude_list is not None:
                    subdir_list = filter(lambda x: all([False if r.search(x) else True for r in self.filter_exclude_list]), subdir_list)
                    file_list = filter(lambda x: all([False if r.search(x) else True for r in self.filter_exclude_list]), file_list)
                # scan folders
                for subdir in subdir_list:
                    full_path = os.path.join(dir_path, subdir)
                    if len(os.listdir(full_path)) == 0:
                        # track empty directories with a dummy entry, non-empty directories should not have entries, they are handled automatically by having files inside them
                        relative_path = os.path.relpath(full_path, self.dir)
                        if self.force_posix_path_sep:
                            relative_path = relative_path.replace(os.path.sep, "/")
                        self.dict_current[relative_path] = {"size": 0, "mtime": 0, "crc": "0", "dir": True}
                        self.set_dirs.add(relative_path)
                # scan files
                for file_name in file_list:
                    full_path = os.path.join(dir_path, file_name)
                    relative_path = os.path.relpath(full_path, self.dir)
                    if self.force_posix_path_sep:
                        relative_path = relative_path.replace(os.path.sep, "/")
                    scan_status.update(relative_path)
                    self.scanFile(full_path, relative_path)
            scan_status.endProgress()
            # check for missing (or moved) files
            for relative_path in (set(self.dict_prev) - set(self.dict_current)):
                if "dir" not in self.dict_prev[relative_path]:
                    if not self.pathMatch(relative_path, self.ignored_toplevel_folders):
                        self.set_missing.add(relative_path)

    def scanFile(self, full_path: str, relative_path: str) -> None:
        # get file attributes and create entry
        size = os.path.getsize(full_path)
        mtime = os.path.getmtime(full_path)
        self.dict_current[relative_path] = {"size": size, "mtime": mtime}
        if self.compare_mode == "crc":
            self.dict_current[relative_path]["crc"] = self.calcCrc(full_path)
        # check if file is new, modified, or corrupted
        if relative_path in self.dict_prev:
            # calculate crc for attr+ unless it's an exact time match and there's a previous crc to copy
            if self.compare_mode == "attr+" and not ("crc" in self.dict_prev[relative_path] and self.fileMatch(relative_path, relative_path, self.dict_prev, set(), exact_time=True)):
                self.dict_current[relative_path]["crc"] = self.calcCrc(full_path)
            # checking if the file changed, accounting for time rounding and DST
            if self.fileMatch(relative_path, relative_path, self.dict_prev, set(), exact_time=False):
                # unchanged file (probably) (keep old crc value if exists and not already recalculated)
                self.set_unmodified.add(relative_path)
                if self.compare_mode in ["attr", "attr+"] and "crc" in self.dict_prev[relative_path] and "crc" not in self.dict_current[relative_path]:
                    self.dict_current[relative_path]["crc"] = self.dict_prev[relative_path]["crc"]
            else:
                # changed file (or corrupted and added to self.set_crc_errors by fileMatch)
                if relative_path not in self.set_crc_errors:
                    self.set_modified.add(relative_path)
        else:
            # new file
            self.set_new.add(relative_path)
            if self.compare_mode == "attr+":
                self.dict_current[relative_path]["crc"] = self.calcCrc(full_path)

    def getMovedAndUpdateLists(self, a_only: list, b_only: list, a_dict: dict, b_dict: dict, compare_func: typing.Callable) -> tuple:
        # f1 is in a is "source" and f2 is in b is "dest"
        moved = []
        for f1 in reversed(a_only):
            if "dir" not in a_dict[f1]:
                for f2 in reversed(b_only):
                    if "dir" not in b_dict[f2]:
                        if compare_func(f1, f2):
                            moved.append({"source": f1, "dest": f2, "match": "source"})
                            a_only.remove(f1)
                            b_only.remove(f2)
                            break
        moved.reverse()
        return moved

    def compareDb(self, other_db: dict, other_crc_errors: set, detect_moves: bool, exact_time: bool, ignore_empty_dirs: bool) -> dict:
        # init and filter file sets (for ignored paths, user filters are done on scan)
        ignored_path_match = lambda f: self.pathMatch(f, self.ignored_toplevel_folders)
        is_dir = lambda d, f: ignore_empty_dirs and "dir" in d[f] and d[f]["dir"] is True
        self_set = set(filter(lambda f: not ignored_path_match(f) and not is_dir(self.dict_current, f), self.dict_current))
        other_set = set(filter(lambda f: not ignored_path_match(f) and not is_dir(other_db, f), other_db))
        # compare file sets
        changed_compare_func = lambda f: not self.fileMatch(f, f, other_db, other_crc_errors, exact_time=exact_time)
        changed = set(filter(changed_compare_func, self_set & other_set))
        changed = sorted(list(changed - (self.set_crc_errors | other_crc_errors)))
        self_only = sorted(list(self_set - other_set))
        other_only = sorted(list(other_set - self_set))
        moved = []
        if detect_moves:
            moved_compare_func = lambda f1, f2: self.fileMatch(f1, f2, other_db, other_crc_errors, exact_time=True)
            moved = self.getMovedAndUpdateLists(self_only, other_only, self.dict_current, other_db, moved_compare_func)
        return {"self_only": self_only, "other_only": other_only, "changed": changed, "moved": moved}

    def compareDirInfo(self, other_info: 'DirInfo', no_moves: bool) -> dict:
        diff = self.compareDb(other_info.dict_current, other_info.set_crc_errors, not no_moves, False, False)
        for pair in diff["moved"]:
            if pair["source"] not in self.set_modified and pair["dest"] not in other_info.set_modified:
                other_info.set_missing.discard(pair["source"])
                self.set_missing.discard(pair["dest"])
        return diff



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

from .dirinfo import DirInfo
from .logman import LogManager
from .utils import getString


class TransferLists:
    def __init__(self, dir_compare: dict):
        """Holds lists of files that are queued for copying/moving/deleting"""
        self.source_only = dir_compare["self_only"]
        self.dest_only = dir_compare["other_only"]
        self.changed = dir_compare["changed"]
        self.moved = dir_compare["moved"]
        self.source_deleted = []
        self.dest_deleted = []

    def __setattr__(self, name, value):
        if not hasattr(self, "locked") or not self.locked[0]:
            super().__setattr__(name, value)
        else:
            raise Exception("Error: Attempted modification to TransferLists after freezing")

    def getLists(self) -> tuple:
        """Returns tuple of lists (or tuples): source_only, dest_only, changed, moved"""
        return (self.source_only,
                self.dest_only,
                self.changed,
                self.moved,
                self.source_deleted,
                self.dest_deleted)

    def getSets(self) -> tuple:
        """Returns tuple of set(lists): source_only, dest_only, changed, moved (files on source), moved (files on dest)"""
        return (set(self.source_only),
                set(self.dest_only),
                set(self.changed),
                set([f["source"] for f in self.moved]),
                set([f["dest"] for f in self.moved]),
                set(self.source_deleted),
                set(self.dest_deleted))

    def isEmpty(self) -> bool:
        return (len(self.source_only) == 0 and
                len(self.dest_only) == 0 and
                len(self.changed) == 0 and
                len(self.moved) == 0 and
                len(self.source_deleted) == 0 and
                len(self.dest_deleted) == 0)

    def freeze(self) -> None:
        """Make this object mostly immutable for code safety"""
        self.source_only = tuple(self.source_only)
        self.dest_only = tuple(self.dest_only)
        self.changed = tuple(self.changed)
        self.moved = tuple(self.moved)
        self.locked = [True]

    def _unfreeze(self) -> None:
        """Allow unfreezing for internal methods (obviously not actually private, but should be treated as such)"""
        self.locked[0] = False
        self.source_only = list(self.source_only)
        self.dest_only = list(self.dest_only)
        self.changed = list(self.changed)
        self.moved = list(self.moved)

    def skipFileTransfers(self, log: LogManager) -> bool:
        self._unfreeze()
        log.append([getString("### SKIPPED ###")], ["Section"])
        print(log.colourString(getString("Enter file paths to remove them from the transfer queues, then 'continue' when ready or 'cancel' to abort"), "G"))
        while True:
            p = input("> ")
            if len(p) == 0 or p == "?":
                print(log.colourString(getString("Enter file paths to remove them from the transfer queues, then 'continue' when ready or 'cancel' to abort"), "G"))
            elif p == "continue":
                self.freeze()
                return True
            elif p == "cancel":
                self.freeze()
                return False
            elif p in self.source_only:
                self.source_only.remove(p)
                log.append(["File:", "Source", p], ["Header", "Subheader", "Path"])
            elif p in self.dest_only:
                self.dest_only.remove(p)
                log.append(["File:", "Dest", p], ["Header", "Subheader", "Path"])
            elif p in self.changed:
                self.changed.remove(p)
                log.append(["File:", "Changed", p], ["Header", "Subheader", "Path"])
            elif p in self.source_deleted:
                self.source_deleted.remove(p)
                log.append(["File:", "Deleted", p], ["Header", "Subheader", "Path"])
            elif p in self.dest_deleted:
                self.dest_deleted.remove(p)
                log.append(["File:", "Deleted", p], ["Header", "Subheader", "Path"])
            else:
                print(log.colourString(getString("Could not find file in queues: %s") % (p), "Y"))

    def propagateSyncDeletions(self, source: DirInfo, dest: DirInfo) -> None:
        source_only, dest_only, _, source_moved, dest_moved, _, _ = self.getSets()
        source_diff = source.compareDb(source.getDatabaseX2(), set(), False, True, False)
        source_new, source_modified, source_missing = set(source_diff["self_only"]), set(source_diff["changed"]), set(source_diff["other_only"])
        dest_diff = dest.compareDb(dest.getDatabaseX2(), set(), False, True, False)
        dest_new, dest_modified, dest_missing = set(dest_diff["self_only"]), set(dest_diff["changed"]), set(dest_diff["other_only"])
        # file was deleted on one side, and should be deleted from the other iff it exists and is not new or modified since the last scan
        # if it was moved on one side, it would have been removed from other_only, this is verified under checkConsistency
        source_deleted = (source_missing & dest_only) - (dest_new | dest_modified | dest_moved)
        dest_deleted = (dest_missing & source_only) - (source_new | source_modified | source_moved)
        self.source_only = sorted(list(source_only - dest_deleted))
        self.source_deleted = sorted(list(source_deleted))
        self.dest_only = sorted(list(dest_only - source_deleted))
        self.dest_deleted = sorted(list(dest_deleted))

    def updateSyncMovedDirection(self, dest: DirInfo) -> None:
        # default action is to leave as (move on dest to) match source if unsure
        dest_diff = dest.compareDb(dest.getDatabaseX2(), set(), False, True, False)
        dest_new, dest_missing = set(dest_diff["self_only"]), set(dest_diff["other_only"])
        for pair in self.moved:
            if pair["dest"] in dest_new and pair["source"] in dest_missing:
                # renamed on dest, should move on source to match dest
                pair["match"] = "dest"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import argparse
import datetime
import os
import sys
import time
import typing

from .config import ConfigObject
from .dirinfo import DirInfo
from .fileman import FileManager
from .logman import LogManager
from .transferlists import TransferLists
from .utils import (
    getString,
    getVersion,
    readJson,
    writeJson,
    testConsistency
)


class BackupManager():
    def __init__(self, args: typing.Union[argparse.Namespace, dict], gui: bool = False):
        """Main class, configure with an argparse namespace or dictionary to create a job then execute with .run()"""
        # init time
        self.backup_time = datetime.datetime.now().strftime("%y%m%d-%H%M")
        # debugging/testing
        if "backup_time_override" in args and args["backup_time_override"]:
            self.backup_time = args["backup_time_override"]
            time.ctime = lambda t: time.asctime(time.gmtime(t))
        # init log manager
        self.log = LogManager(self.backup_time, gui)
        # init config
        if type(args) != dict:
            args = vars(args)
        self.config = ConfigObject(args)
        # load config (be careful if using a non-default config_dir!)
        if "load" in args and args["load"] is True:
            self.loadConfig()
        # update log manager to reference the same config (don't create a new ConfigObject after this point)
        self.log.config = self.config
        # set args that can overwrite loaded config
        if "dry_run" in args and args["dry_run"] is True:
            self.config.dry_run = True
        if "scan_only" in args and args["scan_only"] is True:
            self.config.scan_only = True
        if "compare_mode" in args and args["compare_mode"] is not None and not gui:
            self.config.compare_mode = args["compare_mode"]
        # scan only mode
        if self.config.scan_only and (self.config.dest is None or not os.path.isdir(self.config.dest)):
            self.config.dest = self.config.source
        # check source & dest
        if not os.path.isdir(self.config.source):
            print(self.log.colourString(getString("Invalid source directory: ") + self.config.source, "R"))
            sys.exit()
        if self.config.dest is None:
            print(self.log.colourString(getString("Destination directory not provided or config failed to load"), "R"))
            sys.exit()
        self.config.source = os.path.abspath(self.config.source)
        self.config.dest = os.path.abspath(self.config.dest)
        # save config
        if "save" in args and args["save"] is True:
            self.saveConfig()
        # init gui flag
        self.gui = gui
        # gui modifications
        if self.gui:
            from .gui import simplePrompt
            self.gui_simplePrompt = simplePrompt
            self.config.stdout_status_bar = False
        # log settings
        self.log.append([getString("### SETTINGS ###")])
        self.log.append([getString("Time:"), self.backup_time,
                         getString("Version:"), getVersion(),
                         getString("Source DB CRC:"), "0",
                         getString("Dest DB CRC:"), "0",
                         getString("Config:"), str(vars(self.config))])
        # lock config from future changes (makes code safer and easier to verify)
        self.config.locked = True

    def saveConfig(self) -> None:
        writeJson(os.path.join(self.config.source, self.config.config_dir, "config.json"), vars(self.config))
        print(self.log.colourString(getString("Config saved"), "G"))
        sys.exit()

    def loadConfig(self) -> None:
        current_source = self.config.source
        config_dir = os.path.abspath(os.path.join(self.config.source, self.config.config_dir, "config.json"))
        config = readJson(config_dir)
        print(self.log.colourString(getString("Loaded config from:") + "\n" + config_dir, "G"))
        self.config = ConfigObject(config)
        if self.config.source is None or os.path.abspath(current_source) != os.path.abspath(self.config.source):
            print(self.log.colourString(getString("A config file matching the specified source was not found (case sensitive)"), "R"))
            sys.exit()

    def abortRun(self) -> int:
        self.log.append([getString("### ABORTED ###")])
        self.log.writeLog("database.aborted.json")
        print(self.log.colourString(getString("Run aborted"), "Y"))
        return 1

    def _databaseAndCorruptionCheck(self, dest_database_load_success: bool) -> bool:
        # get databases
        source_dict, source_prev = self.source.getDicts()
        source_new, source_modified, source_missing, source_crc_errors, _, _ = self.source.getSets()
        dest_dict, dest_prev = self.dest.getDicts()
        dest_new, dest_modified, dest_missing, dest_crc_errors, _, _ = self.dest.getSets()
        # print database conflicts, including both collisions from files being modified independently on both sides and unexpected missing files
        # note: this only notifies the user so they can intervene, it does not handle them in any special way, treating them as regular file changes
        # it can also be triggered by time zone or dst changes, lower file system mod time precision, and corruption if using CRCs (handled next)
        abort_run = False
        if dest_database_load_success and self.config.source != self.config.dest:
            self.log.append([getString("### DATABASE CONFLICTS ###")], ["Section"])
            if self.config.main_mode == "sync":
                sync_conflicts = sorted(list(source_modified & dest_modified))  # modified on both sides
                sync_conflicts += sorted(list(source_new & dest_new))  # new on both sides
                sync_conflicts = sorted(list(filter(lambda f: not self.source.fileMatch(f, f, dest_dict, set(), True), sync_conflicts)))  # don't already match
                if len(sync_conflicts) >= 1:
                    print(self.log.colourString(getString("WARNING: found files modified in both source and destination since last scan"), "Y"))
                    abort_run = True
                print(self.log.colourString(getString("Sync Database Conflicts: %s") % (len(sync_conflicts)), "V"))
                self.log.printSyncDbConflicts(sync_conflicts, source_dict, dest_dict, source_prev, dest_prev)
            else:
                dest_conflicts = sorted(list(dest_modified))
                dest_conflicts += sorted(list(dest_missing))
                dest_conflicts += sorted(list(dest_new))
                if len(dest_conflicts) >= 1:
                    print(self.log.colourString(getString("WARNING: found files modified in the destination since last scan"), "Y"))
                    abort_run = True
                print(self.log.colourString(getString("Destination Database Conflicts: %s") % (len(dest_conflicts)), "V"))
                self.log.printChangedFiles(dest_conflicts, dest_dict, dest_prev, "   Dest", "     DB")
        # print database conflicts concerning CRCs if available, as well as CRC conflicts between source and dest if attributes otherwise match
        if len(source_crc_errors) > 0 or len(dest_crc_errors) > 0:
            self.log.append([getString("### CRC ERRORS DETECTED ###")], ["Section"])
            print(self.log.colourString(getString("WARNING: found non matching CRC values, possible corruption detected"), "Y"))
            abort_run = True
            crc_errors_detected = sorted(list(source_crc_errors | dest_crc_errors))
            print(self.log.colourString(getString("CRC Errors Detected: %s") % (len(crc_errors_detected)), "V"))
            if self.config.source != self.config.dest:
                self.log.printSyncDbConflicts(crc_errors_detected, source_dict, dest_dict, source_prev, dest_prev)
            else:
                self.log.printChangedFiles(crc_errors_detected, source_dict, source_prev, " Source", "     DB")
        return abort_run

    def _printAndLogScanOnlyDiffSummary(self, side_str: str, side_info: DirInfo) -> None:
        # get databases
        side_dict, side_prev = side_info.getDicts()
        side_new, side_modified, side_missing, side_crc_errors, side_dirs, side_unmodified = side_info.getSets()
        self_compare = side_info.compareDb(side_prev, set(), detect_moves=True, exact_time=False, ignore_empty_dirs=True)
        list_new, list_missing, list_modified, moved = self_compare["self_only"], self_compare["other_only"], self_compare["changed"], self_compare["moved"]
        # print differences
        sum_size = self.log.prettySize(sum(side_dict[f]["size"] for f in list_new)).strip()
        print(self.log.colourString(getString("%s New Files: %s (%s)") % (side_str, len(list_new), sum_size), "V"))
        self.log.append([getString("### %s NEW FILES ###") % (side_str.upper())], ["Section"])
        self.log.printFiles(list_new, side_dict)
        sum_size = self.log.prettySize(sum(side_prev[f]["size"] for f in list_missing)).strip()
        print(self.log.colourString(getString("%s Missing Files: %s (%s)") % (side_str, len(list_missing), sum_size), "V"))
        self.log.append([getString("### %s MISSING FILES ###") % (side_str.upper())], ["Section"])
        self.log.printFiles(list_missing, side_prev)
        sum_size = self.log.prettySize(sum(abs(side_dict[f]["size"] - side_prev[f]["size"]) for f in list_modified)).strip()
        print(self.log.colourString(getString("%s Changed Files: %s (%s)") % (side_str, len(list_modified), sum_size), "V"))
        self.log.append([getString("### %s CHANGED FILES ###") % (side_str.upper())], ["Section"])
        self.log.printChangedFiles(list_modified, side_dict, side_prev, "    New", "    Old")
        sum_size = self.log.prettySize(sum(side_dict[f["source"]]["size"] for f in moved)).strip()
        print(self.log.colourString(getString("%s Moved Files: %s (%s)") % (side_str, len(moved), sum_size), "V"))
        self.log.append([getString("### %s MOVED FILES ###") % (side_str.upper())], ["Section"])
        self.log.printMovedFiles(moved, side_dict, side_prev, "   New: ", "   Old: ")

    def _printAndLogCompareDiffSummary(self, transfer_lists: TransferLists) -> None:
        # get lists and databases
        source_only, dest_only, changed, moved, source_deleted, dest_deleted = transfer_lists.getLists()
        source_dict, _ = self.source.getDicts()
        dest_dict, _ = self.dest.getDicts()
        # prepare diff messages
        if self.config.noarchive:
            archive_msg = getString("delete")
        else:
            archive_msg = getString("archive")
        if self.config.main_mode == "sync":
            dest_msg = getString("(will be copied to source)")
            move_msg = getString("(will move other file to match)")
        elif self.config.main_mode == "backup":
            dest_msg = getString("(will be left as is)")
            move_msg = getString("(will move files on dest to match source)")
        elif self.config.main_mode == "mirror":
            dest_msg = getString("(will be %sd)" % (archive_msg))
            move_msg = getString("(will move files on dest to match source)")
        if self.config.select_mode == "source":
            change_msg = getString("(%s dest and copy source -> dest)" % (archive_msg))
        elif self.config.select_mode == "dest":
            change_msg = getString("(%s source and copy dest -> source)" % (archive_msg))
        elif self.config.select_mode == "new":
            change_msg = getString("(%s older and copy newer)" % (archive_msg))
        elif self.config.select_mode == "no":
            change_msg = getString("(will be left as is)")
        # print differences
        sum_size = self.log.prettySize(sum(source_dict[f]["size"] for f in source_only)).strip()
        print(self.log.colourString(getString("Source Only (will be copied to dest): %s (%s)") % (len(source_only), sum_size), "V"))
        self.log.append([getString("### SOURCE ONLY ###")], ["Section"])
        self.log.printFiles(source_only, source_dict)
        sum_size = self.log.prettySize(sum(dest_dict[f]["size"] for f in dest_only)).strip()
        print(self.log.colourString(getString("Destination Only %s: %s (%s)") % (dest_msg, len(dest_only), sum_size), "V"))
        self.log.append([getString("### DESTINATION ONLY ###")], ["Section"])
        self.log.printFiles(dest_only, dest_dict)
        sum_size = self.log.prettySize(sum(abs(source_dict[f]["size"] - dest_dict[f]["size"]) for f in changed)).strip()
        print(self.log.colourString(getString("Changed Files %s: %s (%s)") % (change_msg, len(changed), sum_size), "V"))
        self.log.append([getString("### CHANGED FILES ###")], ["Section"])
        self.log.printChangedFiles(changed, source_dict, dest_dict)
        if not self.config.nomoves:
            sum_size = self.log.prettySize(sum(source_dict[f["source"]]["size"] for f in moved)).strip()
            print(self.log.colourString(getString("Moved Files %s: %s (%s)") % (move_msg, len(moved), sum_size), "V"))
            self.log.append([getString("### MOVED FILES ###")], ["Section"])
            self.log.printMovedFiles(moved, source_dict, dest_dict)
        if self.config.main_mode == "sync" and self.config.sync_propagate_deletions:
            sum_size = self.log.prettySize(sum(dest_dict[f]["size"] for f in source_deleted)).strip()
            print(self.log.colourString(getString("Deleted from source (will %s on dest): %s (%s)") % (archive_msg, len(source_deleted), sum_size), "V"))
            self.log.append([getString("### DELETED FROM SOURCE ###")], ["Section"])
            self.log.printFiles(source_deleted, dest_dict)
            sum_size = self.log.prettySize(sum(source_dict[f]["size"] for f in dest_deleted)).strip()
            print(self.log.colourString(getString("Deleted from dest (will %s on source): %s (%s)") % (archive_msg, len(dest_deleted), sum_size), "V"))
            self.log.append([getString("### DELETED FROM DESTINATION ###")], ["Section"])
            self.log.printFiles(dest_deleted, source_dict)

    def _performBackup(self, transfer_lists: TransferLists, simulation_msg: str) -> None:
        # get lists and databases
        source_only, dest_only, changed, moved, source_deleted, dest_deleted = transfer_lists.getLists()
        source_dict, _ = self.source.getDicts()
        dest_dict, _ = self.dest.getDicts()
        # init file manager
        fileman = FileManager(self.config, self.source, self.dest, self.log, self.backup_time, self.gui)
        # perform the backup/mirror/sync
        self.log.append([getString("### START ") + self.config.main_mode.upper() + simulation_msg.upper() + " ###"])
        print(self.log.colourString(getString("Starting ") + self.config.main_mode, "V"))
        if self.config.main_mode == "mirror":
            fileman.handleDeletedFiles(self.config.dest, dest_only)
            fileman.copyFiles(self.config.source, self.config.dest, source_only, source_only)
            fileman.handleMovedFiles(moved)
            fileman.handleChangedFiles(self.config.source, self.config.dest, source_dict, dest_dict, changed)
        elif self.config.main_mode == "backup":
            fileman.copyFiles(self.config.source, self.config.dest, source_only, source_only)
            fileman.handleMovedFiles(moved)
            fileman.handleChangedFiles(self.config.source, self.config.dest, source_dict, dest_dict, changed)
        elif self.config.main_mode == "sync":
            fileman.handleDeletedFiles(self.config.source, dest_deleted)
            fileman.handleDeletedFiles(self.config.dest, source_deleted)
            fileman.copyFiles(self.config.source, self.config.dest, source_only, source_only)
            fileman.copyFiles(self.config.dest, self.config.source, dest_only, dest_only)
            fileman.handleMovedFiles(moved)
            fileman.handleChangedFiles(self.config.source, self.config.dest, source_dict, dest_dict, changed)

    def run(self) -> int:
        """Main method, use this to run your job"""
        # dry run confirmation message
        if self.config.dry_run:
            simulation_msg = getString(" dry run")
            print(self.log.colourString(getString("Dry Run"), "V"))
        else:
            simulation_msg = ""
        # init DirInfo and load previous scan data if available
        self.source = DirInfo(self.config.source, self.config.source_unique_id,
                              self.config.dest, self.config, self.gui)
        self.dest = DirInfo(self.config.dest, self.config.dest_unique_id,
                            self.config.source, self.config, self.gui)
        dest_database_load_success = False
        self.source.loadDatabase()
        self.dest.loadDatabase()
        if self.dest.dict_prev != {}:
            dest_database_load_success = True
        # scan directories (also calculates CRC if enabled) (didn't parallelize scans to prevent excess vibration of adjacent consumer grade disks and keep status bars simple)
        self.log.colourPrint(getString("Scanning files on source:\n%s") % (self.config.source), "B")
        self.source.scanDir(self.config.stdout_status_bar)
        if self.config.source != self.config.dest:
            self.log.colourPrint(getString("Scanning files on destination:\n%s") % (self.config.dest), "B")
            self.dest.scanDir(self.config.stdout_status_bar)
        else:
            self.dest = self.source
        # update log manager to reference the same source and dest
        self.log.source, self.log.dest = self.source, self.dest
        # compare directories (should be relatively fast, all the read operations are done during scan)
        if not self.config.scan_only:
            self.log.colourPrint(getString("Comparing directories..."), "B")
            transfer_lists = TransferLists(self.source.compareDirInfo(self.dest, self.config.nomoves))
            if self.config.main_mode == "sync":
                transfer_lists.updateSyncMovedDirection(self.dest)
                if self.config.sync_propagate_deletions:
                    transfer_lists.propagateSyncDeletions(self.source, self.dest)
            transfer_lists.freeze()
        # check for database conflicts or corruption
        detected_database_conflicts_or_corruption = self._databaseAndCorruptionCheck(dest_database_load_success)
        if self.config.quit_on_db_conflict and detected_database_conflicts_or_corruption:
            return self.abortRun()
        # print differences between current and previous scans then exit if only scanning
        if self.config.scan_only:
            self._printAndLogScanOnlyDiffSummary("Source", self.source)
            if self.config.source != self.config.dest:
                self._printAndLogScanOnlyDiffSummary("Destination", self.dest)
            self.log.append([getString("### SCAN COMPLETED ###")])
            self.log.writeLog("database.json")
            print(self.log.colourString(getString("Completed!"), "G"))
            return 0
        # print differences between source and dest
        self._printAndLogCompareDiffSummary(transfer_lists)
        # consistency checks used for testing and debugging
        if self.backup_time == "000000-0000":
            testConsistency(self.source.getDicts(), self.source.getSets(),
                            self.dest.getDicts(), self.dest.getSets(),
                            transfer_lists.getSets(),
                            self.source.compareDb(self.source.dict_prev, set(), False, False, True),
                            self.dest.compareDb(self.dest.dict_prev, set(), False, False, True),
                            self.source.compareDb(self.source.dict_prev, set(), True, False, True),
                            self.dest.compareDb(self.dest.dict_prev, set(), True, False, True))
        # exit if directories already match
        if transfer_lists.isEmpty():
            print(self.log.colourString(getString("Directories already match, completed!"), "G"))
            self.log.append([getString("### NO CHANGES FOUND ###")])
            self.log.writeLog("database.json")
            return 0
        # wait for go ahead
        self.log.writeLog("database.tmp.json")
        if not self.config.noprompt:
            if self.gui:
                go = self.gui_simplePrompt(getString("Scan complete, continue with %s%s?") % (self.config.main_mode, simulation_msg))
            else:
                print(self.log.colourString(getString("Scan complete, continue with %s%s (y/N/skip)?") % (self.config.main_mode, simulation_msg), "G"))
                go = input("> ")
            if len(go) == 4 and go.lower() == "skip":
                if not transfer_lists.skipFileTransfers(self.log):
                    return self.abortRun()
            elif len(go) == 0 or go[0].lower() != "y":
                return self.abortRun()
        # backup operations
        self._performBackup(transfer_lists, simulation_msg)
        self.log.append([getString("### COMPLETED ###")])
        self.log.writeLog("database.json")
        print(self.log.colourString(getString("Completed!"), "G"))
        return 0



import setuptools
import backupy

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="BackuPy",
    version=backupy.getVersion(),
    description="A simple backup program in python with an emphasis on data integrity and transparent behaviour",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/elesiuta/backupy",
    packages=['backupy'],
    entry_points={
        'console_scripts': [
            'backupy = backupy:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: System :: Archiving :: Backup",
        "Topic :: System :: Archiving :: Mirroring",
        "Topic :: Utilities",
        "Intended Audience :: End Users/Desktop",
        "Environment :: Console",
        "Development Status :: 5 - Production/Stable",
    ],
    test_suite = 'tests.tests',
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import csv
import json
import os


def getVersion() -> str:
    return "1.8.4"


def getString(text: str) -> str:
    # import locale
    # logic for localisation goes here, set language with either a global or singleton
    # store strings in a dictionary or use this as an alias for gettext
    return text


def writeCsv(file_path: str, data: list) -> None:
    if not os.path.isdir(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))
    if os.path.isfile(file_path) and not os.access(file_path, os.W_OK):
        file_path = file_path[:-4] + "-1.csv"
    try:
        with open(file_path, "w", newline="", encoding="utf-8", errors="backslashreplace") as f:
            writer = csv.writer(f, delimiter=",")
            writer.writerows(data)
    except Exception:
        file_path = file_path[:-4] + "-1.csv"
        with open(file_path, "w", newline="", encoding="utf-8", errors="backslashreplace") as f:
            writer = csv.writer(f, delimiter=",")
            writer.writerows(data)


def readJson(file_path: str) -> dict:
    if os.path.exists(file_path):
        with open(file_path, "r", encoding="utf-8", errors="surrogateescape") as json_file:
            data = json.load(json_file)
        return data
    return {}


def writeJson(file_path: str, data: dict, subdir: bool = True, sort_keys: bool = False) -> None:
    if subdir and not os.path.isdir(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))
    try:
        with open(file_path, "w", encoding="utf-8", errors="surrogateescape") as json_file:
            json.dump(data, json_file, indent=1, separators=(',', ': '), sort_keys=sort_keys, ensure_ascii=False)
    except Exception:
        print(getString("Error, could not write: ") + file_path)


def testConsistency(source_dicts: tuple, source_sets: tuple,
                    dest_dicts: tuple, dest_sets: tuple,
                    transfer_lists: tuple,
                    redundant_source: dict, redundant_dest: dict,
                    redundant_source_moves: dict, redundant_dest_moves: dict) -> None:
    # get sets
    source_only, dest_only, changed, source_moved, dest_moved, source_deleted, dest_deleted = transfer_lists
    source_dict, source_prev = source_dicts
    source_dict, source_prev = set(source_dict), set(source_prev)
    source_new, source_modified, source_missing, source_crc_errors, source_dirs, source_unmodified = source_sets
    dest_dict, dest_prev = dest_dicts
    dest_dict, dest_prev = set(dest_dict), set(dest_prev)
    dest_new, dest_modified, dest_missing, dest_crc_errors, dest_dirs, dest_unmodified = dest_sets
    # make sure each item only appears in one set
    union_len = len(source_only | dest_only | changed | source_moved | dest_moved | source_deleted | dest_deleted)
    total_len = len(source_only) + len(dest_only) + len(changed) + len(source_moved) + len(dest_moved) + len(source_deleted) + len(dest_deleted)
    assert union_len == total_len
    # these versions fail because side comparasion can find crc errors and won't remove them from side_x
    # union_len = len(source_new | source_modified | source_missing | source_crc_errors | source_dirs | source_unmodified)
    # total_len = len(source_new) + len(source_modified) + len(source_missing) + len(source_crc_errors) + len(source_dirs) + len(source_unmodified)
    # union_len = len(dest_new | dest_modified | dest_missing | dest_crc_errors | dest_dirs | dest_unmodified)
    # total_len = len(dest_new) + len(dest_modified) + len(dest_missing) + len(dest_crc_errors) + len(dest_dirs) + len(dest_unmodified)
    # contradictions
    assert not (source_moved & dest_moved)
    assert not (source_only & source_moved)
    assert not (dest_only & dest_moved)
    # changed - ... below would be any skipped files
    # assert changed <= (source_modified | dest_modified) | (source_new & dest_new) | (source_crc_errors | dest_crc_errors)
    assert source_only <= source_dict
    assert dest_only <= dest_dict
    assert not source_only & dest_only
    assert not source_only & changed
    assert not dest_only & changed
    # prev dirs and prev files under .backupy cause the next two asserts to be <=
    assert source_dict <= (source_prev - (source_missing | dest_moved)) | source_new | source_dirs
    assert dest_dict <= (dest_prev - (dest_missing | source_moved)) | dest_new | dest_dirs
    assert source_prev >= (source_unmodified | source_missing) - source_new  # same, basically old stuff that should have been ignored in the old dictionary
    assert dest_prev >= (dest_unmodified | dest_missing) - dest_new
    # basically redo all the logic done during dir and file scan using compareDb to see if they match
    assert set(redundant_source["changed"]) == source_modified
    assert set(redundant_source["other_only"]) >= source_missing
    assert set(redundant_source["other_only"]) <= source_missing | dest_moved
    assert set(redundant_source["self_only"]) == source_new
    assert set(redundant_dest["changed"]) == dest_modified
    assert set(redundant_dest["other_only"]) >= dest_missing
    assert set(redundant_dest["other_only"]) <= dest_missing | source_moved
    assert set(redundant_dest["self_only"]) == dest_new
    # no need to check compareDirInfo anymore since it's just a wrapper for compareDb now
    # redundant_dict_compare = self.source.selfCompare(self.dest.dict_current, False, True, False)
    # redundant_dict_compare_reverse = self.dest.selfCompare(self.source.dict_current, False, True, False)
    # assert set(redundant_dict_compare["modified"]) == (changed | (source_crc_errors & source_dict & dest_dict) | (dest_crc_errors & dest_dict & source_dict))
    # assert set(redundant_dict_compare["modified"]) == (changed | (source_crc_errors - (source_dict - dest_dict)) | (dest_crc_errors - (dest_dict - source_dict)))
    # assert set(redundant_dict_compare["new"]) == source_only | source_moved | dest_deleted
    # assert set(redundant_dict_compare["missing"]) == dest_only | dest_moved | source_deleted
    # assert set(redundant_dict_compare_reverse["modified"]) == (changed | (source_crc_errors - (source_dict - dest_dict)) | (dest_crc_errors - (dest_dict - source_dict)))
    # assert set(redundant_dict_compare_reverse["new"]) == dest_only | dest_moved | source_deleted
    # assert set(redundant_dict_compare_reverse["missing"]) == source_only | source_moved | dest_deleted
    # might be able to make a check for moved along the lines of source_only is approx source_new | dest_missing - dest_new - source_missing



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# https://github.com/elesiuta/backupy

import os
import shutil

from .config import ConfigObject
from .dirinfo import DirInfo
from .logman import LogManager
from .statusbar import StatusBar
from .utils import getString


class FileManager:
    # expose the copy function as a class attribute for easy monkey-patching
    copy_function = shutil.copy2

    def __init__(self, config: ConfigObject, source: DirInfo, dest: DirInfo, log: LogManager, backup_time: str, gui: bool):
        """Provides file operation methods (used by BackupManager)"""
        # init variables
        self.log = log
        self.backup_time = backup_time
        self.gui = gui
        self.config = config
        self.source = source
        self.dest = dest

    ##########################################################################
    # Basic file operation methods (only these methods touch files directly) #
    ##########################################################################

    def _removeFile(self, root_path: str, file_relative_path: str) -> None:
        try:
            self.log.append(["Remove:", root_path, file_relative_path])
            if not self.config.dry_run:
                path = os.path.join(root_path, file_relative_path)
                if os.path.isdir(path):
                    try:
                        os.rmdir(path)
                    except IOError:
                        os.chmod(path, 0o777)
                        os.rmdir(path)
                else:
                    try:
                        os.remove(path)
                    except IOError:
                        os.chmod(path, 0o777)
                        os.remove(path)
                if self.config.cleanup_empty_dirs:
                    head = os.path.dirname(path)
                    if len(os.listdir(head)) == 0:
                        os.removedirs(head)
            self.source.updateDictOnRemove(root_path, file_relative_path, self.dest)
        except Exception as e:
            self.log.append(["REMOVE ERROR", root_path, file_relative_path, str(e)])
            print(e)

    def _copyFile(self, source_root: str, dest_root: str, source_file: str, dest_file: str) -> None:
        try:
            self.log.append(["Copy:", source_root, source_file, dest_root, dest_file])
            if not self.config.dry_run:
                source = os.path.join(source_root, source_file)
                dest = os.path.join(dest_root, dest_file)
                if os.path.isdir(source):
                    os.makedirs(dest)
                else:
                    if not os.path.isdir(os.path.dirname(dest)):
                        os.makedirs(os.path.dirname(dest))
                    try:
                        FileManager.copy_function(source, dest)
                    except IOError:
                        os.chmod(dest, 0o777)
                        FileManager.copy_function(source, dest)
                    if self.config.verify_copy:
                        self.source.verifyCrcOnCopy(source_root, dest_root, source_file, dest_file, self.dest)
            self.source.updateDictOnCopy(source_root, dest_root, source_file, dest_file, self.dest)
        except Exception as e:
            self.log.append(["COPY ERROR", source_root, source_file, dest_root, dest_file, str(e)])
            print(e)

    def _moveFile(self, source_root: str, dest_root: str, source_file: str, dest_file: str) -> None:
        try:
            self.log.append(["Move:", source_root, source_file, dest_root, dest_file])
            if not self.config.dry_run:
                source = os.path.join(source_root, source_file)
                dest = os.path.join(dest_root, dest_file)
                if not os.path.isdir(os.path.dirname(dest)):
                    os.makedirs(os.path.dirname(dest))
                shutil.move(source, dest)
                if self.config.cleanup_empty_dirs:
                    head = os.path.dirname(source)
                    if len(os.listdir(head)) == 0:
                        os.removedirs(head)
            self.source.updateDictOnMove(source_root, dest_root, source_file, dest_file, self.dest)
        except Exception as e:
            self.log.append(["MOVE ERROR", source_root, source_file, dest_root, dest_file, str(e)])
            print(e)

    ##########################################################################
    # Batch file operation methods (do not perform file operations directly) #
    ##########################################################################

    def _removeFiles(self, root_path: str, file_relative_paths: list) -> None:
        if not file_relative_paths:
            return None
        self.log.colourPrint(getString("Removing %s unique files from:\n%s") % (len(file_relative_paths), root_path), "B")
        for f in file_relative_paths:
            self._removeFile(root_path, f)
        self.log.colourPrint(getString("Removal completed!"), "NONE")

    def copyFiles(self, source_root: str, dest_root: str, source_files: list, dest_files: list) -> None:
        if not source_files:
            return None
        self.log.colourPrint(getString("Copying %s unique files from:\n%s\nto:\n%s") % (len(source_files), source_root, dest_root), "B")
        copy_status = StatusBar("Copying", len(source_files), self.config.stdout_status_bar, gui=self.gui)
        for i in range(len(source_files)):
            copy_status.update(source_files[i])
            self._copyFile(source_root, dest_root, source_files[i], dest_files[i])
        copy_status.endProgress()

    def _recycleFiles(self, source_root: str, dest_root: str, source_files: list, dest_files: list) -> None:
        if not source_files:
            return None
        self.log.colourPrint(getString("Archiving %s unique files from:\n%s") % (len(source_files), source_root), "B")
        for i in range(len(source_files)):
            self._moveFile(source_root, dest_root, source_files[i], dest_files[i])
        self.log.colourPrint(getString("Archiving completed!"), "NONE")

    def handleDeletedFiles(self, root_path: str, file_relative_paths: list) -> None:
        if not file_relative_paths:
            return None
        if self.config.noarchive:
            self._removeFiles(root_path, file_relative_paths)
        else:
            recycle_bin = os.path.join(root_path, self.config.trash_dir, self.backup_time)
            self._recycleFiles(root_path, recycle_bin, file_relative_paths, file_relative_paths)

    def handleMovedFiles(self, moved_pairs: list) -> None:
        if moved_pairs and not self.config.nomoves:
            # conflicts shouldn't happen since moved is a subset of files from source_only and dest_only
            # depends on source_info.dirCompare(dest_info) otherwise source and dest keys will be reversed
            self.log.colourPrint(getString("Moving %s files to match both sides") % (len(moved_pairs)), "B")
            for f in moved_pairs:
                if f["match"] == "dest":
                    side = self.config.source
                    oldLoc = f["source"]
                    newLoc = f["dest"]
                elif f["match"] == "source":
                    side = self.config.dest
                    oldLoc = f["dest"]
                    newLoc = f["source"]
                self._moveFile(side, side, oldLoc, newLoc)
            self.log.colourPrint(getString("Moving completed!"), "NONE")

    def _archiveFile(self, root_path: str, file_relative_path: str) -> None:
        if not self.config.noarchive:
            archive_path = os.path.join(root_path, self.config.archive_dir, self.backup_time)
            self._moveFile(root_path, archive_path, file_relative_path, file_relative_path)

    def handleChangedFiles(self, source_root: str, dest_root: str, source_dict: dict, dest_dict: dict, changed: list) -> None:
        if not changed:
            return None
        self.log.colourPrint(getString("Handling %s file changes per selection mode") % (len(changed)), "B")
        copy_status = StatusBar("Copying", len(changed), self.config.stdout_status_bar, gui=self.gui)
        for frp in changed:
            copy_status.update(frp)
            if self.config.select_mode == "source":
                self._archiveFile(dest_root, frp)
                self._copyFile(source_root, dest_root, frp, frp)
            elif self.config.select_mode == "dest":
                self._archiveFile(source_root, frp)
                self._copyFile(dest_root, source_root, frp, frp)
            elif self.config.select_mode == "new":
                if source_dict[frp]["mtime"] > dest_dict[frp]["mtime"]:
                    self._archiveFile(dest_root, frp)
                    self._copyFile(source_root, dest_root, frp, frp)
                else:
                    self._archiveFile(source_root, frp)
                    self._copyFile(dest_root, source_root, frp, frp)
            else:
                break
        copy_status.endProgress()
